@@ -28,7 +28,7 @@ typedef struct {
   adns_rrtype number;
 } AdnsTclRRTypeInfo;
 
-const IdDataSpec adnstcl_queries, adnstcl_resolvers;
+extern const IdDataSpec cht_adnstcl_queries, cht_adnstcl_resolvers;
 
 #include "adns+tcmdif.h"
 
