@@ -20,6 +20,8 @@
 
 #include "chiark_tcl_cdb.h"
 
+#define KEYLEN_MAX (INT_MAX/2)
+
 #define ftello ftell
 #define fseeko fseek
 
@@ -44,7 +46,8 @@ typedef struct Pathbuf {
 #define MAX_SUFFIX 5
 
 static void pathbuf_init(Pathbuf *pb, const char *pathb) {
-  int l= strlen(pathb);
+  size_t l= strlen(pathb);
+  assert(l < INT_MAX);
   pb->buf= TALLOC(l + MAX_SUFFIX + 1);
   memcpy(pb->buf, pathb, l);
   pb->sfx= pb->buf + l;
@@ -160,20 +163,24 @@ typedef struct Rw {
   int ix, autocompact;
   int cdb_fd, lock_fd;
   struct cdb cdb; /* valid iff cdb_fd >= 0 */
-  FILE *logfile;
+  FILE *logfile; /* may be 0; if so, is broken */
   HashTable logincore;
   Pathbuf pbsome, pbother;
   off_t mainsz;
   ScriptToInvoke on_info, on_lexminval;
 } Rw;
 
+static void rw_cdb_close(Tcl_Interp *ip, Rw *rw) {
+  if (rw->cdb_fd >= 0) cdb_free(&rw->cdb);
+  maybe_close(rw->cdb_fd);
+}
+
 static int rw_close(Tcl_Interp *ip, Rw *rw) {
   int rc, r;
 
   rc= TCL_OK;
   ht_destroy(&rw->logincore);
-  if (rw->cdb_fd >= 0) cdb_free(&rw->cdb);
-  maybe_close(rw->cdb_fd);
+  rw_cdb_close(ip,rw);
   maybe_close(rw->lock_fd);
 
   if (rw->logfile) {
@@ -250,7 +257,7 @@ static int readlognum(FILE *f, int delim, int *num_r) {
   *p= 0;
 
   errno=0; ul= strtoul(numbuf, &ep, 10);
-  if (*ep || errno || ul >= INT_MAX/2) return -2;
+  if (*ep || errno || ul >= KEYLEN_MAX) return -2;
   *num_r= ul;
   return 0;
 }
@@ -313,7 +320,7 @@ static int readstorelogrecord(FILE *f, HashTable *ht,
 static int writerecord(FILE *f, const char *key, const HashValue *val) {
   int r;
 
-  r= fprintf(f, "+%zu,%d:%s->", strlen(key), val->len, key);
+  r= fprintf(f, "+%d,%d:%s->", (int)strlen(key), val->len, key);
   if (r<0) return -1;
   
   r= fwrite(val->data, 1, val->len, f);
@@ -411,7 +418,7 @@ static int infocb(Tcl_Interp *ip, Rw *rw, const char *arg1,
 /*---------- Opening ----------*/
 
 static int cdbinit(Tcl_Interp *ip, Rw *rw) {
-  /* On entry, cdb_fd >=0 but cdb is _undefined_/
+  /* On entry, cdb_fd >=0 but cdb is _undefined_
    * On exit, either cdb_fd<0 or cdb is initialised */
   int r, rc;
   
@@ -608,9 +615,9 @@ static int compact_core(Tcl_Interp *ip, Rw *rw, unsigned long logsz,
   a.reccount= reccount_r;
 
   r= fclose(rw->logfile);
+  rw->logfile= 0;
   if (r) { rc= cht_posixerr(ip, errno, "probable data loss!  failed to fclose"
 			    " logfile during compact");  goto x_rc; }
-  rw->logfile= 0;
   
   rc= infocb(ip, rw, "compact-start", "log=%luby main=%luby",
 	     logsz, (unsigned long)rw->mainsz);
@@ -698,7 +705,7 @@ static int compact_core(Tcl_Interp *ip, Rw *rw, unsigned long logsz,
   if (!a.mainfile) PE("create .tmp for new main during compact");
 
   r= ht_forall(&rw->logincore, addto_main, &a);
-  if (r) { rc= cht_posixerr(ip, r, "error writing to new .main"
+  if (r) { rc= cht_posixerr(ip, errno, "error writing to new .main"
 			    " during compact");  goto x_rc; }
 
   r= putc('\n', a.mainfile);
@@ -810,8 +817,7 @@ static int compact_keepopen(Tcl_Interp *ip, Rw *rw, int force) {
 
   rc= compact_core(ip, rw, logsz, &reccount);  if (rc) goto x_rc;
 
-  maybe_close(rw->cdb_fd);
-  rw->cdb_fd= -1;
+  rw_cdb_close(ip,rw);
   ht_destroy(&rw->logincore);
   ht_setup(&rw->logincore);
 
@@ -857,12 +863,22 @@ int cht_do_cdbwr_compact_auto(ClientData cd, Tcl_Interp *ip, void *rw_v) {
 static int update(Tcl_Interp *ip, Rw *rw, const char *key,
 		  const Byte *data, int dlen) {
   HashValue *val;
+  const char *failed;
   int rc, r;
+  off_t recstart;
+
+  if (strlen(key) >= KEYLEN_MAX)
+    return cht_staticerr(ip, "key too long", "CDB KEYOVERFLOW");
 
   if (!rw->logfile) return cht_staticerr
-    (ip, "previous compact failed; cdbwr must be closed and reopened "
-     "before any further updates", "CDB BROKEN");
+    (ip, "failure during previous compact or error recovery;"
+     " cdbwr must be closed and reopened before any further updates",
+     "CDB BROKEN");
   
+  recstart= ftello(rw->logfile);
+  if (recstart < 0)
+    return cht_posixerr(ip, errno, "failed to ftello .jrn during update");
+
   val= htv_prep(dlen);  assert(val);
   memcpy(htv_fillptr(val), data, dlen);
 
@@ -877,6 +893,33 @@ static int update(Tcl_Interp *ip, Rw *rw, const char *key,
 
  x_rc:
   TFREE(val);
+  assert(rc);
+
+  /* Now, we have to try to sort out the journal so that it's
+   * truncated and positioned to where this abortively-written record
+   * started, with no buffered output and the error indicator clear.
+   *
+   * There seems to be no portable way to ensure the buffered unwritten
+   * output is discarded, so we close and reopen the stream.
+   */
+  fclose(rw->logfile);
+
+  rw->logfile= fopen(pathbuf_sfx(&rw->pbsome,".jrn"), "r+");
+  if (!rw->logfile) { failed= "fopen"; goto reset_fail; }
+
+  r= ftruncate(fileno(rw->logfile), recstart);
+  if (r) { failed= "ftruncate"; goto reset_fail; }
+
+  r= fseeko(rw->logfile, recstart, SEEK_SET);
+  if (r) { failed= "fseeko"; goto reset_fail; }
+
+  return rc;
+
+ reset_fail:
+  Tcl_AppendResult(ip, " (additionally, ", failed, " failed"
+		   " in error recovery: ", strerror(errno), ")", (char*)0);
+  if (rw->logfile) { fclose(rw->logfile); rw->logfile= 0; }
+
   return rc;
 }  
 
@@ -908,11 +951,18 @@ static int lookup_rw(Tcl_Interp *ip, void *rw_v, const char *key,
   val= ht_lookup(&rw->logincore, key);
   if (val) {
     if (val->len) { *data_r= val->data; *len_r= val->len; return TCL_OK; }
-    else { *data_r= 0; *len_r= -1; return TCL_OK; }
+    else goto not_found;
   }
 
+  if (rw->cdb_fd<0) goto not_found;
+
   return cht_cdb_lookup_cdb(ip, &rw->cdb, key, strlen(key), data_r, len_r);
-} 
+
+ not_found:
+  *data_r= 0;
+  *len_r= -1;
+  return TCL_OK;
+}
 
 int cht_do_cdbwr_lookup(ClientData cd, Tcl_Interp *ip, void *rw_v,
 			const char *key, Tcl_Obj *def,
